import wrapt

from falken_trace import config
from falken_trace.tracing.datadog import wrap_dd_span
from falken_trace.web.fastapi import wrap_fastapi_entrypoint_span

if config.env_vars_config.falken_trace_enabled:
    if config.env_vars_config.dd_trace_enabled:
        try:
            wrapt.wrap_function_wrapper("ddtrace.trace", "Tracer.start_span", wrap_dd_span)
        except AttributeError:
            # fallback to v2 import
            wrapt.wrap_function_wrapper("ddtrace", "Tracer.start_span", wrap_dd_span)
    if config.env_vars_config.fastapi_installed:
        wrapt.wrap_function_wrapper("fastapi.routing", "run_endpoint_function", wrap_fastapi_entrypoint_span)



from __future__ import annotations

import inspect
import os
from typing import TYPE_CHECKING, Any, Callable

from typing_extensions import ParamSpec

from falken_trace.config import env_vars_config
from falken_trace.utils import get_outer_frames, normalize_path

if TYPE_CHECKING:
    from ddtrace.trace import Span, Tracer

P = ParamSpec("P")


def wrap_dd_span(wrapped: Callable[P, Span], _instance: Tracer, args: Any, kwargs: Any) -> Span:  # noqa: ANN401
    span: Span = wrapped(*args, **kwargs)
    if not env_vars_config.falken_trace_enabled:
        return span

    work_dir = os.getcwd()
    for frame_info in get_outer_frames(inspect.currentframe()):
        if frame_info.function == "func_wrapper" and "ddtrace" in frame_info.filename:  # noqa: SIM102
            # trying to get the actual definition of the callable
            if (func := frame_info.frame.f_locals.get("f")) or (func := frame_info.frame.f_locals.get("coro")):
                file_path = inspect.getsourcefile(func) or ""
                if file_path.startswith(work_dir) and "site-packages" not in file_path:
                    _, lineno = inspect.getsourcelines(func)
                    span.set_tag("code.filepath", normalize_path(path=file_path, path_prefix=work_dir))
                    span.set_tag("code.lineno", str(lineno + 1))  # points originally to the `tracer` decorator
                    span.set_tag("code.func", func.__name__)

        if frame_info.filename.startswith(work_dir) and "site-packages" not in frame_info.filename:
            span.set_tag("code.wrap_filepath", normalize_path(path=frame_info.filename, path_prefix=work_dir))
            span.set_tag("code.wrap_lineno", str(frame_info.lineno))
            span.set_tag("code.wrap_func", frame_info.function)
            break

    return span



from __future__ import annotations

import inspect
import os
from typing import TYPE_CHECKING, Any, Callable, TypeVar

from typing_extensions import ParamSpec

from falken_trace.config import env_vars_config
from falken_trace.utils import flatten_dict, normalize_path

if TYPE_CHECKING:
    from collections.abc import Awaitable

    from ddtrace.trace import Span

if env_vars_config.dd_trace_enabled:
    try:
        from ddtrace.trace import tracer
    except ImportError:
        # keeping it to support lower `ddtrace` versions
        from ddtrace import tracer  # type:ignore[attr-defined]

T = TypeVar("T")
P = ParamSpec("P")

# an API route can be set for both sync and async functions
FUNCTION_PREFIXES = ("async def ", "def ")


async def wrap_fastapi_entrypoint_span(
    wrapped: Callable[P, Awaitable[T]],
    _instance: None,
    args: Any,  # noqa: ANN401
    kwargs: Any,  # noqa: ANN401
) -> T:
    if (dependant := kwargs.get("dependant")) and hasattr(dependant, "call"):
        work_dir = os.getcwd()
        file_path = inspect.getsourcefile(dependant.call) or ""
        lines, lineno = inspect.getsourcelines(dependant.call)

        # `getsourcelines()` returns the code lines with decorators, line breaks, etc.
        line: str | tuple[int, str] | None = next(
            ((i, line.lstrip()) for (i, line) in enumerate(lines) if line.lstrip().startswith(FUNCTION_PREFIXES)), None
        )
        if isinstance(line, tuple):
            lineno += line[0]
            line = line[1].split("(", maxsplit=1)[0]
            for prefix in FUNCTION_PREFIXES:
                line = line.removeprefix(prefix)

        if env_vars_config.dd_trace_enabled and (span := tracer.current_span()):
            # can be `None`, if DD is disabled or FastAPI is not patched
            span.set_tag("code.filepath", normalize_path(path=file_path, path_prefix=work_dir))
            span.set_tag("code.lineno", str(lineno))
            span.set_tag("code.func", line)

            if env_vars_config.falken_include_params and (values := kwargs.get("values")):
                attach_func_params(span=span, param_args=values)

    if inspect.iscoroutinefunction(wrapped):
        return await wrapped(*args, **kwargs)
    else:
        return wrapped(*args, **kwargs)  # type:ignore[return-value]


def attach_func_params(span: Span, param_args: dict[str, Any]) -> None:
    for param, arg in param_args.items():
        if hasattr(arg, "model_dump"):  # true, if `pydantic.BaseModel`
            flat = flatten_dict(arg.model_dump())
            for flat_key, value in flat.items():
                span.set_tag(f"params.{param}.{flat_key}", value)
        else:
            span.set_tag(f"params.{param}", arg)



from __future__ import annotations

import inspect
import os
from importlib.util import find_spec
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Iterator
    from types import FrameType


def get_outer_frames(frame: FrameType | None, max_frames: int = 5) -> Iterator[inspect.FrameInfo]:
    """Get a records for a frame and all higher (calling) frames.

    Copied and adjusted version of inspect.getouterframes()
    """
    frame_count = 0
    while frame and frame_count < max_frames:
        frame_info = (frame,) + inspect.getframeinfo(frame, 0)  # noqa: RUF005
        if frame_count:  # the first frame is itself
            yield inspect.FrameInfo(*frame_info)
        frame = frame.f_back
        frame_count += 1


def normalize_path(path: str, path_prefix: str) -> str:
    return path.removeprefix(path_prefix).lstrip("/")


def flatten_dict(data: dict[str, Any], parent_key: str = "") -> dict[str, str]:
    """Creates a single level dict by concatenating keys with dots and values as strings."""
    items = {}
    for key, value in data.items():
        new_key = f"{parent_key}.{key}" if parent_key else key
        if isinstance(value, dict):
            items.update(flatten_dict(value, new_key))
        else:
            items[new_key] = str(value)
    return items


def is_dd_trace_enabled() -> bool:
    if find_spec("ddtrace") is None:
        return False

    return os.environ.get("DD_TRACE_ENABLED", "true").lower() in ("true", "1")



import os
from importlib.util import find_spec

from falken_trace.utils import is_dd_trace_enabled


class EnvVarsConfig:
    def __init__(self) -> None:
        self.dd_trace_enabled = is_dd_trace_enabled()
        self.falken_trace_enabled = os.environ.get("FALKEN_TRACE_ENABLED", "true").lower() in ("true", "1")
        self.falken_include_params = os.environ.get("FALKEN_INCLUDE_PARAMS", "true").lower() in ("true", "1")
        self.fastapi_installed = bool(find_spec("fastapi"))


env_vars_config = EnvVarsConfig()
